#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include "kexec.h"


struct memory_range *memory_range;
static int memory_ranges;

int valid_memory_range(struct kexec_segment *segment)
{
	int i;
	for(i = 0; i < memory_ranges; i++) {
		unsigned long mstart, mend;
		unsigned long sstart, send;
		mstart = memory_range[i].start;
		mend = memory_range[i].end;
		sstart = (unsigned long)segment->mem;
		send = sstart + segment->memsz -1;
		/* Check to see if we are fully contained */
		if ((mstart <= sstart) && (mend >= send)) {
			return 0;
		}
	}
	return -1;
}

int sort_segments(struct kexec_segment *segment, int segments)
{
	int i, j;
	void *end;
	/* Do a stupid insertion sort... */
	for(i = 1; i < segments; i++) {
		int temp_idx;
		struct kexec_segment temp;
		temp_idx = i;
		for(j = i +1; j < segments; j++) {
			if (segment[j].mem < segment[temp_idx].mem) {
				temp_idx = j;
			}
		}
		if (temp_idx != i) {
			temp = segment[temp_idx];
			segment[temp_idx] = segment[i];
			segment[i] = temp;
		}
	}
	/* Now see if any of the segments overlap */
	end = 0;
	for(i = 1; i < segments; i++) {
		if (end > segment[i].mem) {
			fprintf(stderr, "Overlapping memory segments at %p\n",
				end);
			return -1;
		}
		end = ((char *)segment[i].mem) + segment[i].memsz;
	}
	return 0;
}

int locate_arguments(struct kexec_segment *segment, int segments,
	size_t align, unsigned long max)
{
	int i, j;
	struct memory_range *mem_range;
	int max_mem_ranges, mem_ranges;

	/* Compute the free memory ranges */
	max_mem_ranges = memory_ranges + (segments -1);
	mem_range = malloc(max_mem_ranges *sizeof(struct memory_range));
	mem_ranges = 0;
		
	/* Perform a merge on the 2 sorted lists of memory ranges  */
	for(j = 1, i = 0; i < memory_ranges; i++) {
		unsigned long sstart, send;
		unsigned long mstart, mend;
		mstart = memory_range[i].start;
		mend = memory_range[i].end;
		while((j < segments) && (((unsigned long)segment[j].mem) <= mend)) {
			sstart = (unsigned long)segment[j].mem;
			send = sstart + segment[j].memsz -1;
			if (mstart < sstart) {
				mem_range[mem_ranges].start = mstart;
				mem_range[mem_ranges].end = sstart -1;
				mem_ranges++;
			}
			mstart = send +1;
			j++;
		}
		if (mstart <= mend) {
			mem_range[mem_ranges].start = mstart;
			mem_range[mem_ranges].end = mend;
			mem_ranges++;
		}
	}
	/* Now find the first memory_range I can use */
	for(i = 0; i < mem_ranges; i++) {
		unsigned long start, size;
		start = (mem_range[i].start + align -1) & ~(align -1);
		size = mem_range[i].end - start;
		if (size >= segment[0].memsz) {
			break;
		}
	}
	if (i == mem_ranges) {
		fprintf(stderr, "Could not find a free area of memory for the arguemnts...\n");
		return -1;
	}
	if (mem_range[i].end > max) {
		fprintf(stderr, "Could not find a free area of memory below: %lx...\n",
			max);
		return -1;
	}
	segment[0].mem = (void *)((mem_range[i].start + align - 1) & ~(align -1));
	return 0;
}

static char *cmd_line_str(int argc, char **argv)
{
	int command_line_len;
	char *buff;
	int i;
	command_line_len = 0;
	for(i = 0; i < argc; i++) {
		command_line_len += strlen(argv[i]) + 1;
	}
	buff = malloc(command_line_len);
	buff[0] = '\0';
	for(i = 0; i < argc; i++) {
		if (i != 0) {
			strcat(buff, " ");
		}
		strcat(buff, argv[i]);
	}
	return buff;
}

int main(int argc, char *argv[])
{
	char *kernel;
	char *command_line;
	FILE *fp_kernel;
	int i, type;
	struct kexec_segment *segments;
	int nr_segments;
	void *entry;
	if (argc <= 1) {
		fprintf(stderr, "No kernel specified\n");
		return -1;
	}
	kernel = argv[1];
	command_line = cmd_line_str(argc -2, argv +2);
	
	fp_kernel = fopen(kernel, "rb");
	if (!fp_kernel) {
		fprintf(stderr, "Cannot open %s: %s\n",
			kernel, strerror(errno));
		return -1;
	}
	if (get_memory_ranges(&memory_range, &memory_ranges) < 0) {
		fprintf(stderr, "Could not get memory layout\n");
		return -1;
	}
	
	for(i = 0; i < file_types; i++) {
		if (file_type[i].probe(fp_kernel) > 0) {
			break;
		}
	}
	if (i == file_types) {
		fprintf(stderr, "Cannot determine the file type of %s\n",
			kernel);
		return -1;
	}
	type = i;
	if (file_type[type].load(fp_kernel, command_line,
		    &entry, &segments, &nr_segments) < 0) {
		fprintf(stderr, "Cannot load %s\n", kernel);
		return -1;
	}
	/* Verify all of the segments load to a valid location in memory */
	for(i = 1; i < nr_segments; i++) {
		if (valid_memory_range(segments +i) < 0) {
			fprintf(stderr, "Invalid memory segment %p - %p\n",
				segments[i].mem,
				((char *)segments[i].mem) + segments[i].memsz);
			return -1;
		}
	}
	/* Sort the segments and verify we don't have overlaps */
	if (sort_segments(segments, nr_segments) < 0) {
		return -1;
	}
	kexec(entry, nr_segments, segments);
	/* I have failed if I make it here */
	fprintf(stderr, "kexec failed: %s\n",
		strerror(errno));
	fprintf(stderr, "entry       = %p\n", entry);
	fprintf(stderr, "nr_segments = %d\n", nr_segments);
	for(i = 0; i < nr_segments; i++) {
		fprintf(stderr, "segment[%d].buf   = %p\n", i, segments[i].buf);
		fprintf(stderr, "segment[%d].bufsz = %x\n", i, segments[i].bufsz);
		fprintf(stderr, "segment[%d].mem   = %p\n", i, segments[i].mem);
		fprintf(stderr, "segment[%d].memsz = %x\n", i, segments[i].memsz);
	}
	return -1;
} 


